use std::sync::Arc;

use tokio::sync::Mutex;
use futures::{
    select,
    FutureExt,
};


use hexland_server::{
    GlobalState,
    Channel,
    channel_pair,
    GameController,
    message::{Message, MessageWebSocket},
    msg,
};


pub struct Client {
    global_state: Arc<GlobalState>,
    ws: MessageWebSocket,
    channel: Option<Channel>,
}

impl Client {
    pub fn new(ws: MessageWebSocket, global_state: Arc<GlobalState>) -> Self {
        Client {
            global_state,
            ws,
            channel: None
        }
    }
    pub async fn run(&mut self) {
        loop {
            match &mut self.channel {
                Some(channel) => {
                    select! {
                        msg = channel.rx.recv().fuse() =>
                            self.handle_server_msg(msg.unwrap()).await,
                        msg = self.ws.next().fuse() =>
                            self.handle_client_msg(msg.unwrap()).await,
                    };
                }
                None => {
                    let msg = self.ws.next().await;
                    self.handle_client_msg(msg.unwrap()).await;
                }
            }
        }
    }
    async fn handle_client_msg(&mut self, msg: Message) {
        match msg.command.as_str() {
            "HOST" => {
                let room_code = self.global_state.code_generator.lock().await.generate();
                let mut game_controller = GameController::default();

                let (client_channel, server_channel) = channel_pair();
                self.channel = Some(client_channel);
                game_controller.channels.push(server_channel);

                let game_controller = Arc::new(Mutex::new(game_controller));
                self.global_state.rooms.lock().await.insert(room_code.clone(), Arc::clone(&game_controller));
                tokio::spawn(async move {game_loop(game_controller)});
                self.ws.send(msg!(ROOM_CODE, room_code)).await.unwrap();
            }
            "JOIN" => {
                let room_code = &msg.args[0];
                let rooms = self.global_state.rooms.lock().await;
                let room = rooms.get(room_code);

                match room {
                    Some(room) => {
                        let mut room = room.lock().await;
                        let (client_channel, server_channel) = channel_pair();
                        self.channel = Some(client_channel);
                        room.channels.push(server_channel);
                        self.ws.send(msg!(JOIN_OK)).await.unwrap();
                    }
                    None => {
                        self.ws.send(msg!(JOIN_INVALID)).await.unwrap();
                    }
                }
            }
            _ => if let Some(channel) = &self.channel {
                // Forward message to the server
                channel.tx.send(msg).await.unwrap();
            }
        }
    }
    async fn handle_server_msg(&mut self, _msg: Message) {
        todo!();
    }
}

fn game_loop(_gc: Arc<Mutex<GameController>>) {
    todo!();
}
