use std::convert::{From, TryFrom};

use tokio::net::TcpStream;
use tokio_tungstenite::{
    WebSocketStream,
    tungstenite::Message as WsMessage,
};
use futures::{StreamExt, SinkExt};

#[derive(PartialEq, Debug)]
pub struct Message {
    pub command: String,
    pub args: Vec<String>,
}

pub type Result<T> = std::result::Result<T, Error>;

#[repr(u8)]
#[derive(PartialEq, Debug)]
pub enum Error {
    BadParse,
    NonText,
    Unknown,
}

impl Message {
    pub fn parse(text: String) -> Result<Message> {
	let re = regex::Regex::new(r"^([A-Z_]+):\s*(.*)").unwrap();
	match re.captures(text.as_str()) {
            Some(captures) => {
		if captures.len() < 3 {
		    Err(Error::BadParse)
		} else {
		    let command = captures.get(1).unwrap().as_str().to_string();
		    let args = captures.get(2).unwrap().as_str()
			.split(',')
			.map(|s| s.trim().to_string())
			.collect();
                    Ok(Message { command, args })
                }
            }
	    None => Err(Error::BadParse),
        }
    }
}

impl TryFrom<WsMessage> for Message {
    type Error = Error;

    fn try_from(ws_message: WsMessage) -> Result<Self> {
        let text = match ws_message {
            WsMessage::Text(text) => text,
            _ => return Err(Error::NonText)
        };
        Message::parse(text)
    }
}

impl From<Message> for WsMessage {
    fn from(message: Message) -> Self {
        WsMessage::Text(message.to_string())
    }
}

pub struct MessageWebSocket(pub WebSocketStream<TcpStream>);

impl MessageWebSocket {
    pub async fn next(&mut self) -> Result<Message> {
        if let Some(Ok(msg)) = self.0.next().await {
            msg.try_into()
        } else {
            Err(Error::Unknown)
        }
    }
    pub async fn send(&mut self, msg: Message) -> Result<()> {
        if self.0.send(msg.into()).await.is_err() {
            Err(Error::Unknown)
        } else {
            Ok(())
        }
    }
}

#[macro_export]
macro_rules! msg {
    ( $command:ident) => {
	{
	    let command = stringify!($command).to_string();
	    let args = vec![];
	    Message { command, args }
	}
    };
    ( $command:ident, $( $arg:expr ),*) => {
	{
	    let command = stringify!($command).to_string();
	    let args = vec![$($arg.to_string()),*];
	    Message { command, args }
	}
    };
}
	    

impl std::fmt::Display for Message {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::result::Result<(), std::fmt::Error> {
        if self.args.is_empty() {
            write!(f, "{}:", self.command)
        } else {
	    write!(f, "{}: {}", self.command, self.args.as_slice().join(", "))
        }
    }
}

#[cfg(test)]
mod test {
    use super::*;
    #[test]
    fn test_parse() -> Result<()> {
	let text = "COMMAND: arg1, arg2";
	let msg = Message::parse(text.to_string())?;
	assert_eq!(msg!(COMMAND, "arg1", "arg2"), msg);
	Ok(())
    }
    #[test]
    fn test_to_string() {
        let msg = msg!(COMMAND, "arg1", "arg2");
	assert_eq!(msg.to_string(), "COMMAND: arg1, arg2".to_string());
    }
}
